/*
  Benchmark of signal mixing implementations
  (with and without coefficient / "gain"), comparison between:

    1) pure C++ implementation
    2) assembly SSE implementation (Ardour)
    3) utilization of GCC vector extension

  Adjust the Makefile's CFLAGS for your machine before compiling!

  (C) 2008, Christian Schoenebeck <cuse@users.sf.net>
*/

#include <math.h>
#include <time.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>

#include "mix.h"

#define FRAGMENTSIZE	1000
#define RUNS            100000
#define PROOF_RESULT    false
#define ALLOC_BUFFERS   0

#if ALLOC_BUFFERS
float* pSampleInputBuf __attribute__ ((aligned (16))); // just a buffer with random data
float* pOutput __attribute__ ((aligned (16)));
#else
float pSampleInputBuf[FRAGMENTSIZE] __attribute__ ((aligned (16))); // just a buffer with random data
float pOutput[FRAGMENTSIZE] __attribute__ ((aligned (16)));
#endif

typedef float* __attribute__ ((aligned (16))) audio_buf;
typedef float v4sf __attribute__ ((vector_size(16)));

void gcc_vector_mix(const audio_buf pSrcBuf, audio_buf pDstBuf, const int Samples) {
    const v4sf* src = static_cast<v4sf*>((void*)pSrcBuf);
    v4sf* dst       = static_cast<v4sf*>((void*)pDstBuf);
    const int cells = Samples / 4;
    for (int i = 0; i < cells; ++i)
        dst[i] += src[i];
}

void gcc_vector_mix(const audio_buf pSrcBuf, audio_buf pDstBuf, const int Samples, const float coeff) {
    const v4sf vcoeff = { coeff, coeff, coeff, coeff };
    const v4sf* src = static_cast<v4sf*>((void*)pSrcBuf);
    v4sf* dst       = static_cast<v4sf*>((void*)pDstBuf);
    const int cells = Samples / 4;
    for (int i = 0; i < cells; ++i)
        dst[i] += src[i] * vcoeff;
}

void cpp_mix(const audio_buf pSrcBuf, audio_buf pDstBuf, const int Samples) {
    for (int i = 0; i < Samples; ++i)
        pDstBuf[i] += pSrcBuf[i];
}

void cpp_mix(const audio_buf pSrcBuf, audio_buf pDstBuf, const int Samples, const float fLevel) {
    for (int i = 0; i < Samples; ++i)
        pDstBuf[i] += pSrcBuf[i] * fLevel;
}

// just print the first 15 samples on the console, so we can manually compare the results
void proofResult() {
    if (!PROOF_RESULT) return;
    for (int i = 0; i < 15; ++i)
        printf("%f,", pOutput[i]);
    printf("\n");
}

int main() {
#if ALLOC_BUFFERS
    pSampleInputBuf = (float*) memalign(16,FRAGMENTSIZE*sizeof(float));
    pOutput         = (float*) memalign(16,FRAGMENTSIZE*sizeof(float));
#endif

    // prepare some input data for simulation
    for (int i = 0; i < FRAGMENTSIZE; ++i) {
        pSampleInputBuf[i] = (float)i;
    }

    printf("Benchmarking mixdown (no coeff):\n");

    {
        // zero out output buffers
        memset(pOutput,0,FRAGMENTSIZE*sizeof(float));

        printf("pure C++");

        clock_t stop_time;
        clock_t start_time = clock();

        for (int i = 0; i < RUNS; i++) {
            cpp_mix(pSampleInputBuf, pOutput, FRAGMENTSIZE);
        }

        stop_time = clock();
        float elapsed_time = (stop_time - start_time) / (double(CLOCKS_PER_SEC) / 1000.0);
        printf("\t\t: %1.0f ms\n", elapsed_time);

        proofResult();
    }

    {
        // zero out output buffers
        memset(pOutput,0,FRAGMENTSIZE*sizeof(float));

        printf("ASM SSE");

        clock_t stop_time;
        clock_t start_time = clock();

        for (int i = 0; i < RUNS; i++) {
            x86_sse_mix_buffers_no_gain(pOutput, pSampleInputBuf, FRAGMENTSIZE);
        }

        stop_time = clock();
        float elapsed_time = (stop_time - start_time) / (double(CLOCKS_PER_SEC) / 1000.0);
        printf("\t\t\t: %1.0f ms\n", elapsed_time);

        proofResult();
    }

    {
        // zero out output buffers
        memset(pOutput,0,FRAGMENTSIZE*sizeof(float));

        printf("GCC vector extensions");

        clock_t stop_time;
        clock_t start_time = clock();

        for (int i = 0; i < RUNS; i++) {
            gcc_vector_mix(pSampleInputBuf, pOutput, FRAGMENTSIZE);
        }

        stop_time = clock();
        float elapsed_time = (stop_time - start_time) / (double(CLOCKS_PER_SEC) / 1000.0);
        printf("\t: %1.0f ms\n", elapsed_time);

        proofResult();
    }



    printf("\nBenchmarking mixdown (WITH coeff):\n");
    const float coeff = 2.0f; // random "gain" coefficient for mix

    {
        // zero out output buffers
        memset(pOutput,0,FRAGMENTSIZE*sizeof(float));

        printf("pure C++");

        clock_t stop_time;
        clock_t start_time = clock();

        for (int i = 0; i < RUNS; i++) {
            cpp_mix(pSampleInputBuf, pOutput, FRAGMENTSIZE, coeff);
        }

        stop_time = clock();
        float elapsed_time = (stop_time - start_time) / (double(CLOCKS_PER_SEC) / 1000.0);
        printf("\t\t: %1.0f ms\n", elapsed_time);

        proofResult();
    }

    {
        // zero out output buffers
        memset(pOutput,0,FRAGMENTSIZE*sizeof(float));

        printf("ASM SSE");

        clock_t stop_time;
        clock_t start_time = clock();

        for (int i = 0; i < RUNS; i++) {
            x86_sse_mix_buffers_with_gain(pOutput, pSampleInputBuf, FRAGMENTSIZE, coeff);
        }

        stop_time = clock();
        float elapsed_time = (stop_time - start_time) / (double(CLOCKS_PER_SEC) / 1000.0);
        printf("\t\t\t: %1.0f ms\n", elapsed_time);

        proofResult();
    }

    {
        // zero out output buffers
        memset(pOutput,0,FRAGMENTSIZE*sizeof(float));

        printf("GCC vector extensions");

        clock_t stop_time;
        clock_t start_time = clock();

        for (int i = 0; i < RUNS; i++) {
            gcc_vector_mix(pSampleInputBuf, pOutput, FRAGMENTSIZE, coeff);
        }

        stop_time = clock();
        float elapsed_time = (stop_time - start_time) / (double(CLOCKS_PER_SEC) / 1000.0);
        printf("\t: %1.0f ms\n", elapsed_time);

        proofResult();
    }
}
